package t1ck.merger

import java.io. {

  BufferedReader,
  BufferedWriter,
  InputStreamReader,
  File,
  FileInputStream,
  FileReader,
  FileWriter,
  OutputStreamWriter,
  Writer

}

import java.util.zip. { GZIPInputStream }

import net.liftweb.common. { Loggable, Logger }

import org.apache.commons.cli. {

  HelpFormatter,
  Option,
  OptionBuilder,
  Options,
  ParseException,
  PosixParser

}

import scala.collection.mutable.ArrayBuffer
import scala.util.control.Breaks. { break, breakable }

/** Sort and merge multiple files into stdout.
 *
 * @author myst3r10n
 */
class Main extends xsbti.AppMain with Logger with Loggable {

	protected class Exit(val code: Int) extends xsbti.Exit

	def run(configuration: xsbti.AppConfiguration) = {

    // Needed: https://github.com/harrah/xsbt/issues/435#issuecomment-6681951
    Thread.currentThread.setContextClassLoader(getClass.getClassLoader)

    Main.main(configuration.arguments)

    new Exit(0)

	}
}

/** Companion.
 *
 * Run:
 *
 * run usr/csv/a.csv usr/csv/b.csv usr/csv/c.csv
 *
 * @author myst3r10n
 */
object Main extends Logger with Loggable {

  def main(args: Array[String]) {

    val options = new Options {

      OptionBuilder.withArgName("delimiter")
      OptionBuilder.hasArg
      OptionBuilder.withDescription("delimiter of fields")
      addOption(OptionBuilder.create('d'))

      addOption(new Option("h", "print this message"))

      OptionBuilder.withArgName("records")
      OptionBuilder.hasArg
      OptionBuilder.withDescription("ignore initial records")
      addOption(OptionBuilder.create('i'))

      OptionBuilder.withArgName("record")
      OptionBuilder.hasArg
      OptionBuilder.withDescription("markt label record")
      addOption(OptionBuilder.create('l'))

      OptionBuilder.withArgName("column")
      OptionBuilder.hasArg
      OptionBuilder.withDescription("column to sort")
      addOption(OptionBuilder.create('s'))

    }

    var delimiter = ","
    var ignoreRecords = 1
    var labelRecord = 1
    var sortColumn = 0
    var writer = new BufferedWriter(new OutputStreamWriter(System.out))

    val files = ArrayBuffer[File]()

    try {

      val line = (new PosixParser).parse(options, args)

      if(line.hasOption('d'))
        delimiter = line.getOptionValue('d')

      if(line.hasOption('h')) {

        (new HelpFormatter).printHelp("t1ck-merger", options)
        sys.exit(0)

      }

      if(line.hasOption('i'))
        ignoreRecords = line.getOptionValue('i').toInt

      if(line.hasOption('l'))
        labelRecord = line.getOptionValue('l').toInt

      if(line.hasOption('s'))
        sortColumn = line.getOptionValue('s').toInt

      if(!line.getArgs.isEmpty)
        line.getArgs.foreach(files += new File(_))

      val reader = new BufferedReader(new InputStreamReader(System.in))

      while(reader.ready)
        breakable {

          val line = reader.readLine

          if(line == null)
            break

          if(line.contains('\0'))
            line.split('\0').foreach(file => files += new File(file))

        }

    } catch {

      case e: ParseException =>
        error(e.getMessage)
        sys.exit(1)

    }

    val readers = ArrayBuffer[BufferedReader]()

    try {

      files.toArray.foreach { file =>

        if(file.getName.toLowerCase.endsWith(".gz"))
          readers += new BufferedReader(new InputStreamReader(
            new GZIPInputStream(new FileInputStream(file))))
        else
          readers += new BufferedReader(new FileReader(file))

      }

      var current = -1

      // Ignore records and read label
      current = 0
      var labeled = false
      readers.foreach { reader =>

        breakable {

          for(r <- 1 to ignoreRecords) {

            var line = reader.readLine

            if(line == null) {

              warn("Skipped: " + files(current).getName)

              readers.remove(current)
              files.remove(current)
              current -= 1

              break

            }

            if(!labeled && r == labelRecord) {

              labeled = true

              writer.write(line)
              writer.newLine
              writer.flush

            }
          }
        }

        current += 1

      }

      val records = ArrayBuffer[Array[String]]()

      // Read first record and split into fields
      current = 0
      readers.foreach { reader =>

        var line = reader.readLine

        if(line == null || line.isEmpty) {

          warn("Skipped: " + files(current).getName)

          readers.remove(current)
          files.remove(current)
          current -= 1

        } else
          records += line.split(delimiter)

        current += 1

      }

      // Sort and merge records into single stdout
      var lowestFile = -1
      while(!readers.isEmpty) {

        var i = 0
        records.foreach { record =>

          if(
            lowestFile == -1 ||
            records(lowestFile)(sortColumn) > record(sortColumn))
            lowestFile = i

          i += 1

        }

        writer.write(records(lowestFile).reduceLeft(_ + delimiter + _))
        writer.newLine
        writer.flush

        var line = readers(lowestFile).readLine

        if(line == null || line.isEmpty) {

          info("Done: " + files(lowestFile).getName)

          readers.remove(lowestFile)
          files.remove(lowestFile)
          records.remove(lowestFile)
          lowestFile = -1

        } else
          records(lowestFile) = line.split(delimiter)

      }
    } finally {

      readers.foreach(_.close)

    }
  }
}

